// compiler avec 
// gcc -g -Wall  -D_REENTRANT -D_GNU_SOURCE lecteur_redacteur.c
//                                -lm -lpthread -o lectrect
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>

#include <pthread.h>
#include <errno.h>
#include <string.h>

/***************************************/
/*                                     */
/*             entete                  */
/*                                     */
/***************************************/
typedef struct {
  int num;
  int seed;
  int nbtours;
} param_t;

typedef struct {
  int nb_lect;
  int nb_red;

  pthread_mutex_t mut;
  pthread_mutex_t mut_ordre;

  pthread_cond_t attente_lect;
  pthread_cond_t attente_red;
} lect_red_t;

lect_red_t* lect_red_init();
void lect_red_destroy(lect_red_t *m);
void lect_red_demande_lecture(lect_red_t *m);
void lect_red_fin_lecture(lect_red_t *m);
void lect_red_demande_ecriture(lect_red_t *m);
void lect_red_fin_ecriture(lect_red_t *m);

/* #include "lecteur_redacteur.h" */

/***************************************/
/*                                     */
/*               code                  */
/*                                     */
/***************************************/
lect_red_t *m;

lect_red_t* lect_red_init() {
  int res;
  lect_red_t *m = malloc(sizeof(lect_red_t));

  // Initialisation des compteurs à 0
  m->nb_lect = 0;
  m->nb_red = 0;

  // Initialisation des mutex sans attributs
  res = pthread_mutex_init(&(m->mut), NULL);
  if (res != 0) {
    fprintf(stderr, "lect_red_init :  %s", strerror(res));
    exit(1);
  }
  res = pthread_mutex_init(&(m->mut_ordre), NULL);
  if (res != 0) {
    fprintf(stderr, "lect_red_init :  %s", strerror(res));
    exit(1);
  }


  // Initialisation des variables conditionnelles
  res = pthread_cond_init(&(m->attente_lect), NULL);
  if (res != 0) {
    fprintf(stderr, "lect_red_init :  %s", strerror(res));
    exit(1);
  }
  res = pthread_cond_init(&(m->attente_red), NULL);
  if (res != 0) {
    fprintf(stderr, "lect_red_init :  %s", strerror(res));
    exit(1);
  }

  return m;
}

void lect_red_destroy(lect_red_t *m) {
  int res;

  // Récupération des ressources des mutex
  res = pthread_mutex_destroy(&(m->mut));
  if (res != 0) {
    fprintf(stderr, "lect_red_destroy :  %s", strerror(res));
    exit(1);
  }
  res = pthread_mutex_destroy(&(m->mut_ordre));
  if (res != 0) {
    fprintf(stderr, "lect_red_destroy :  %s", strerror(res));
    exit(1);
  }

  res = pthread_cond_destroy(&(m->attente_lect));
  if (res != 0) {
    fprintf(stderr, "lect_red_destroy :  %s", strerror(res));
    exit(1);
  }

  res = pthread_cond_destroy(&(m->attente_red));
  if (res != 0) {
    fprintf(stderr, "lect_red_destroy :  %s", strerror(res));
    exit(1);
  }
  
  free(m);
}

void lect_red_demande_lecture(lect_red_t *m) {
  int res;


  res = pthread_mutex_lock(&(m->mut_ordre));
  if (res != 0) {
    fprintf(stderr, "lect_red_demande_lecture :  %s", strerror(res));
    exit(1);
  }
  res = pthread_mutex_lock(&(m->mut));
  if (res != 0) {
    fprintf(stderr, "lect_red_demande_lecture :  %s", strerror(res));
    exit(1);
  }

  while (m->nb_red != 0) {
    // la boucle while sert à  retester l'attente en sortant
    // Elle est nécessaire car les redacteur réveillent systématiquement
    // les lecteur et les rédacteurs
    pthread_cond_wait(&(m->attente_lect), &(m->mut));
  }
  m->nb_lect ++;

  res = pthread_mutex_unlock(&(m->mut));
  if (res != 0) {
    fprintf(stderr, "lect_red_demande_lecture :  %s", strerror(res));
    exit(1);
  }
  res = pthread_mutex_unlock(&(m->mut_ordre));
  if (res != 0) {
    fprintf(stderr, "lect_red_demande_lecture :  %s", strerror(res));
    exit(1);
  }
}

void lect_red_fin_lecture(lect_red_t *m) {
  int res;

  res = pthread_mutex_lock(&(m->mut));
  if (res != 0) {
    fprintf(stderr, "lect_red_fin_lecture :  %s", strerror(res));
    exit(1);
  }
  m->nb_lect --;
  if (m->nb_lect == 0) {
    // C'est le dernier thread en lecture, on réveille les rédacteurs en 
    // attente.
    pthread_cond_signal(&(m->attente_red));
  }
  res = pthread_mutex_unlock(&(m->mut));
  if (res != 0) {
    fprintf(stderr, "lect_red_fin_lecture :  %s", strerror(res));
    exit(1);
  }
}

void lect_red_demande_ecriture(lect_red_t *m) {
  int res;


  res = pthread_mutex_lock(&(m->mut_ordre));
  if (res != 0) {
    fprintf(stderr, "lect_red_demande_ecriture :  %s", strerror(res));
    exit(1);
  }
  res = pthread_mutex_lock(&(m->mut));
  if (res != 0) {
    fprintf(stderr, "lect_red_demande_ecriture :  %s", strerror(res));
    exit(1);
  }
  
  while (m->nb_red != 0 || m->nb_lect != 0) {
     pthread_cond_wait(&(m->attente_red), &(m->mut));
  }

  m->nb_red ++;

  res = pthread_mutex_unlock(&(m->mut));
  if (res != 0) {
    fprintf(stderr, "lect_red_demande_ecriture :  %s", strerror(res));
    exit(1);
  }
  res = pthread_mutex_unlock(&(m->mut_ordre));
  if (res != 0) {
    fprintf(stderr, "lect_red_demande_ecriture :  %s", strerror(res));
    exit(1);
  }

}

void lect_red_fin_ecriture(lect_red_t *m) {
  int res;

  res = pthread_mutex_lock(&(m->mut));
  if (res != 0) {
    fprintf(stderr, "lect_red_fin_ecriture :  %s", strerror(res));
    exit(1);
  }
 
  m->nb_red --;
  pthread_cond_broadcast(&(m->attente_lect));
  pthread_cond_signal(&(m->attente_red));

  res = pthread_mutex_unlock(&(m->mut));
  if (res != 0) {
    fprintf(stderr, "lect_red_fin_ecriture :  %s", strerror(res));
    exit(1);
  }

}

void *lecture(param_t *a) {
  int i;
  srand(a->seed);

  for (i=0;i<a->nbtours; i++) {
    lect_red_demande_lecture(m);
    printf("rlecteur \tnum %d\ttour %d : en section critique %d\tlect %d\tred\n", a->num, i, m->nb_lect, m->nb_red);
    usleep((int) ceil(rand()*100.0/RAND_MAX*1.0));
    lect_red_fin_lecture(m);
    usleep((int) ceil(rand()*100.0/RAND_MAX*1.0));    
  }
  
  return NULL;
}

void *ecriture(param_t *a) {
  int i;
  srand(a->seed);

  for (i=0;i<a->nbtours; i++) {
    lect_red_demande_ecriture(m);
    printf("redacteur \tnum %d\ttour %d : en section critique %d\tlect %d\tred\n", a->num, i, m->nb_lect, m->nb_red);
    usleep((int) ceil(rand()*100.0/RAND_MAX*1.0));
    lect_red_fin_ecriture(m);
    usleep((int) ceil(rand()*100.0/RAND_MAX*1.0));    
  }
  
  return NULL;
}

int main(int argc, char *argv[]) {
  int nbl, nbr, nbt, i;
  int res;

  if (
      (argc != 4)
      || ((nbl=atoi(argv[1]))==0)
      || ((nbr=atoi(argv[2]))==0)
      || ((nbt=atoi(argv[3]))==0)
      )
    {
      fprintf(stderr, "usage %s <nblect> <nbred> <nbtours>\n", argv[0]);
      exit(1);
    }
  
  // tableau qui contiendra les arguement au threads
  param_t args[nbl+nbr];
  // tableua qui contiendra les numéro de threads (pour fair un join)
  pthread_t tab[nbl+nbr];

  m = lect_red_init();
  
  for (i=0; i<nbl; i++) {
    args[i].num=i;
    args[i].seed = rand();
    args[i].nbtours = nbt;
    res  = pthread_create(&tab[i], NULL,(void * (*)(void *)) lecture, &args[i]);
    if (res) {
      fprintf(stderr, "create !! %s\n", strerror(res));
      exit(1);
    }
  }
  for (i=0; i<nbr; i++) {
    args[i+nbl].num=i;
    args[i+nbl].seed = rand();
    args[i].nbtours = nbt;
    res  = pthread_create(&tab[i+nbl], NULL,(void * (*)(void *)) ecriture, &args[i]);
    if (res) {
      fprintf(stderr, "create !! %s\n", strerror(res));
      exit(1);
    }
  }
  

  for (i=0; i<nbl+nbr; i++) {
    res = pthread_join(tab[i], NULL);
    if (res) {
      fprintf(stderr, "join !! %s\n", strerror(res));
      exit(1);
    }
  }

  return 0;
}





