// compiler avec
// gcc -g -Wall -pedantic -D_REENTRANT -D_GNU_SOURCE lecteur_redacteur.c -lm -lpthread -o lectrect

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>

#include <pthread.h>

/***************************************/
/*                                     */
/*             entete                  */
/*                                     */
/***************************************/

typedef struct LectRed {
  int nb_lect;
  int nb_red;
  pthread_mutex_t mut;
  pthread_mutex_t mut_ordre;

  pthread_cond_t attente_lect;
  pthread_cond_t attente_red;
} LectRed;



LectRed* LR_init();
void LR_destroy(LectRed *m);
void LR_demande_lecture(LectRed *m);
void LR_fin_lecture(LectRed *m);
void LR_demande_ecriture(LectRed *m);
void LR_fin_ecriture(LectRed *m);

/* #include "lecteur_redacteur.h" */

/***************************************/
/*                                     */
/*               code                  */
/*                                     */
/***************************************/
LectRed *m;

LectRed* LR_init() {
  int res;
  LectRed *m = malloc(sizeof(LectRed));

  // Initialisation des compteurs à 0
  m->nb_lect = 0;
  m->nb_red = 0;

  // Initialisation des mutex sans attributs
  res = pthread_mutex_init(&(m->mut), NULL);
  if (res != 0) {
    perror("LR_init : ");
    exit(1);
  }
  res = pthread_mutex_init(&(m->mut_ordre), NULL);
  if (res != 0) {
    perror("LR_init : ");
    exit(1);
  }


  // Initialisation des variables conditionnelles
  res = pthread_cond_init(&(m->attente_lect), NULL);
  if (res != 0) {
    perror("LR_init : ");
    exit(1);
  }
  res = pthread_cond_init(&(m->attente_red), NULL);
  if (res != 0) {
    perror("LR_init : ");
    exit(1);
  }

  return m;
}

void LR_destroy(LectRed *m) {
  int res;

  // Récupération des ressources des mutex
  res = pthread_mutex_destroy(&(m->mut));
  if (res != 0) {
    perror("LR_destroy : ");
    exit(1);
  }
  res = pthread_mutex_destroy(&(m->mut_ordre));
  if (res != 0) {
    perror("LR_destroy : ");
    exit(1);
  }

  res = pthread_cond_destroy(&(m->attente_lect));
  if (res != 0) {
    perror("LR_destroy : ");
    exit(1);
  }

  res = pthread_cond_destroy(&(m->attente_red));
  if (res != 0) {
    perror("LR_destroy : ");
    exit(1);
  }
}

void LR_demande_lecture(LectRed *m) {
  int res;

  res = pthread_mutex_lock(&(m->mut));
  if (res != 0) {
    perror("LR_demande_lecture : ");
    exit(1);
  }

  res = pthread_mutex_lock(&(m->mut_ordre));
  if (res != 0) {
    perror("LR_demande_lecture : ");
    exit(1);
  }

  while (m->nb_red != 0) {
    // la boucle while sert à  retester l'attente en sortant
    // Elle est nécessaire car les redacteur réveillent systématiquement
    // les lecteur et les rédacteurs
    pthread_cond_wait(&(m->attente_lect), &(m->mut));
  }
  m->nb_lect ++;

  res = pthread_mutex_unlock(&(m->mut_ordre));
  if (res != 0) {
    perror("LR_demande_lecture : ");
    exit(1);
  }


  res = pthread_mutex_unlock(&(m->mut));
  if (res != 0) {
    perror("LR_demande_lecture : ");
    exit(1);
  }


}
void LR_fin_lecture(LectRed *m) {
  int res;

  res = pthread_mutex_lock(&(m->mut));
  if (res != 0) {
    perror("LR_fin_lecture : ");
    exit(1);
  }
  m->nb_lect --;
  if (m->nb_lect == 0) {
    // C'est le dernier thread en lecture, on réveille les rédacteurs en 
    // attente.
    pthread_cond_signal(&(m->attente_red));
  }
  res = pthread_mutex_unlock(&(m->mut));
  if (res != 0) {
    perror("LR_fin_lecture : ");
    exit(1);
  }
}

void LR_demande_ecriture(LectRed *m) {
  int res;

  res = pthread_mutex_lock(&(m->mut));
  if (res != 0) {
    perror("LR_demande_ecriture : ");
    exit(1);
  }

  res = pthread_mutex_lock(&(m->mut_ordre));
  if (res != 0) {
    perror("LR_demande_ecriture : ");
    exit(1);
  }
  
  while (m->nb_red != 0 || m->nb_lect != 0) {
     pthread_cond_wait(&(m->attente_red), &(m->mut));
  }

  m->nb_red ++;

  res = pthread_mutex_unlock(&(m->mut_ordre));
  if (res != 0) {
    perror("LR_demande_ecriture : ");
    exit(1);
  }

  res = pthread_mutex_unlock(&(m->mut));
  if (res != 0) {
    perror("LR_demande_ecriture : ");
    exit(1);
  }


}


void LR_fin_ecriture(LectRed *m) {
  int res;

  res = pthread_mutex_lock(&(m->mut));
  if (res != 0) {
    perror("LR_fin_ecriture : ");
    exit(1);
  }
 
  m->nb_red --;
  pthread_cond_broadcast(&(m->attente_lect));
  pthread_cond_signal(&(m->attente_red));

  res = pthread_mutex_unlock(&(m->mut));
  if (res != 0) {
    perror("LR_fin_ecriture : ");
    exit(1);
  }

}

void lecture(int seed) {
  srand(seed);

  while (1) {
    LR_demande_lecture(m);
    printf("lecteur : %d lect %d red\n", m->nb_lect, m->nb_red);
    usleep((int) ceil(rand()*100.0/RAND_MAX*1.0));
    LR_fin_lecture(m);
    usleep((int) ceil(rand()*100.0/RAND_MAX*1.0));    
  }
  
}

void ecriture(int seed) {
  srand(seed);

  while (1) {
    LR_demande_ecriture(m);
    printf("redacteur : %d lect %d red\n", m->nb_lect, m->nb_red);
    usleep((int) ceil(rand()*100.0/RAND_MAX*1.0));
    LR_fin_ecriture(m);
    usleep((int) ceil(rand()*100.0/RAND_MAX*1.0));    
  }
  
}

int main(int argc, char *argv[]) {
  int nbl, nbr, i;
  pthread_t *tab = NULL;
  int res;
  int seed;

  if (
      (argc != 3)
      || ((nbl=atoi(argv[1]))==0)
      || ((nbr=atoi(argv[2]))==0)
      )
    {
      fprintf(stderr, "usage %s <nblect> <nbred>\n", argv[0]);
    exit(1);
  }

  m = LR_init();

  tab = malloc((nbl+nbr)*sizeof(pthread_t));

  for (i=0; i<nbl; i++) {
    seed = rand();
    res  = pthread_create(&tab[i], NULL,(void * (*)(void *)) lecture, (void*)seed);
  }
  for (i=0; i<nbr; i++) {
    seed = rand();
    res  = pthread_create(&tab[i+nbl], NULL,(void * (*)(void *)) ecriture, (void*)seed);
  }
  

  while (1);

  return 0;
}





